/**********************************************************************/
/*                                                                    */
/* io.h: LISPME input/output subsystem definitions                    */
/*                                                                    */
/* LispMe System (c) FBI Fred Bayer Informatics                       */
/*                                                                    */
/* Distributed under the GNU General Public License;                  */
/* see the README file. This code comes with NO WARRANTY.             */
/*                                                                    */
/* Modification history                                               */
/*                                                                    */
/* When?      What?                                              Who? */
/* -------------------------------------------------------------------*/
/* 14.06.1997 New                                                FBI  */
/* 25.10.1999 Prepared for GPL release                           FBI  */
/* 01.04.2000 Prepared for GCC 2.0 and SDK 3.5                   FBI  */
/*                                                                    */
/**********************************************************************/

#ifndef INC_IO_H
#define INC_IO_H

/**********************************************************************/
/* Includes                                                           */
/**********************************************************************/
#include "store.h"
#include "port.h"

/**********************************************************************/
/* Several buffer sizes                                               */
/**********************************************************************/
#define MEMO_OUTPUT_SIZE    4097
#define MAX_TOKEN_LEN        512

/**********************************************************************/
/* Printing flags                                                     */
/**********************************************************************/
#define PRT_OUTFIELD 0x00
#define PRT_MESSAGE  0x01
#define PRT_MEMO     0x02
#define PRT_PORT     0x03
#define PRT_DEST     0x03
#define PRT_ESCAPE   0x04
#define PRT_AUTOLF   0x08
#define PRT_SPACE    0x10

/**********************************************************************/
/* Printing foreign types                                             */
/**********************************************************************/
typedef void (*printerType)(Boolean, void*);

/**********************************************************************/
/* Exported data                                                      */
/**********************************************************************/
extern char     msg[];
extern UInt8    token[];
extern Int32    outpSize;
extern Int16    lineNr;
extern Boolean  onlyWS;
extern char*    startPtr;
extern char*    currPtr;
extern char*    (*EOFHandler)(void);
extern char*    limit;
extern OutPort* gOut;

extern OutPort* portOutField;
extern OutPort  portMsg;
extern OutPort  portConv;

/**********************************************************************/
/* Prototypes                                                         */
/**********************************************************************/
void    printSEXP(PTR p, UInt8 flags, OutPort* outp)            SEC(IO);
PTR     readNthSEXP(char* src, int n)                           SEC(IO);
PTR     readBufSEXP(char* src, char* end)                       SEC(IO);
PTR     loadMemo(char* src)                                     SEC(IO);
Boolean isSymbolChar(char c)                                    SEC(IO);
void    registerPrinter(UInt8 ftype, printerType fp)            SEC(IO);
char*   printString(PTR str)                                    SEC(IO);

/**********************************************************************/
/* Helper functions for outputting user defined types                 */
/**********************************************************************/
void    writeSEXP(PTR p)                                        SEC(IO);

#endif
